from .conf import get_default_language
from .translator import get_i18n_field
from .utils import get_language


class ActiveLanguageMixin(object):
    """
    Add this mixin to your admin class to exclude all virtual fields, except:

     - The original field (for the default language, settings.LANGUAGE_CODE)
     - The field for the currently active language, without fallback.
    """

    def get_exclude(self, request, obj=None):
        i18n_field = get_i18n_field(self.model)
        # use default implementation for models without i18n-field
        if i18n_field is None:
            return super().get_exclude(request)

        language = get_language()
        if language == get_default_language():
            language = False

        excludes = []
        for field in i18n_field.get_translated_fields():
            # Not excluded:
            # - language is None: the _i18n-version of the field.
            # - language equals the current language
            if field.language == language:
                continue

            excludes.append(field.name)

        # de-duplicate
        return list(set(excludes))



#!/usr/bin/env python

import os
import sys
from subprocess import STDOUT, CalledProcessError, check_output

MODELS_PY = "migrate_test/app/models.py"


def main():
    """
    This script runs a migration procedure from django-modeltranslation to
    django-moeltrans.

    `pre_migrate_tests` run before the migration to django-modeltrans
    `post_migrate_tests` run after the migration to django-modeltrans
    """

    # change directory to the test_migrations folder
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    print("Current working directory:", os.getcwd())

    # clean up the test projects directory
    cmd("git clean migrate_test/ -f")
    cmd("git checkout -- migrate_test/")

    cmd("pip install -r requirements.txt")

    # start with an empty db
    cmd("dropdb modeltrans-migration --if-exists")

    if "TRAVIS" in os.environ:
        cmd("createdb modeltrans-migration -U postgres")
    else:
        cmd("createdb modeltrans-migration")

    # populate database and do some pre-migration verifications
    manage("migrate")
    manage("loaddata data.json")
    run_test("pre_migrate_tests")

    # do the actual migration modeltranslation -> modeltrans.

    # 1. install django-modeltrans and add to installed apps.
    cmd("pip install -U ..")
    replace_in_file("migrate_test/settings.py", '# "modeltrans"', '"modeltrans"')

    # 2. Uncomment modeltrans i18n-field in models.py
    replace_in_file(MODELS_PY, "# from", "from")
    replace_in_file(MODELS_PY, "# i18n", "i18n")
    replace_in_file(MODELS_PY, "# indexes", "indexes")

    # 3. make the migrations to add django-modeltrans json fields
    manage("makemigrations app")
    manage("migrate app")

    # 4. Create the data migration
    manage("i18n_makemigrations app")
    manage("migrate app")

    # 5. remove django-modeltranslation
    replace_in_file("migrate_test/settings.py", '"modeltranslation",')
    cmd("rm -r migrate_test/app/translation.py")
    replace_in_file(MODELS_PY, ", virtual_fields=False")

    # 6. migrate once more to remove django-modeltranslation's fields
    manage("makemigrations app")
    manage("migrate app")

    # 6. run the post-migration tests
    run_test("post_migrate_tests")


def replace_in_file(file, search, dest=""):
    with open(file, "r") as f:
        contents = f.read()

    with open(file, "w") as f:
        f.write(contents.replace(search, dest))


def cmd(c):
    print("\033[92m Running command: \033[0m", c)

    try:
        result = check_output(c, shell=True, stderr=STDOUT)
        if len(result) > 0:
            print(result.decode().replace("\\n", "\n"))
        return result
    except CalledProcessError as e:
        print("\033[31m Process errored: \033[0m, code: {}".format(e.returncode))
        print(e.output.decode().replace("\\n", "\n"))
        sys.exit(1)


def manage(c):
    cmd("coverage run -a --rcfile=../.coveragerc ./manage.py {}".format(c))


def run_test(test_module):
    assert not test_module.endswith(".py")

    manage("test --keepdb {}".format(test_module))


if __name__ == "__main__":
    main()



from django.apps import apps
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db.models import Manager

from .conf import get_available_languages, get_default_language
from .fields import TranslationField, translated_field_factory
from .manager import MultilingualManager, transform_translatable_fields
from .utils import get_model_field


def get_i18n_field(Model):
    """
    Return the i18n field if the model has it, else None.
    """
    try:
        i18n_field = Model._meta.get_field("i18n")
    except FieldDoesNotExist:
        return

    if not isinstance(i18n_field, TranslationField):
        return

    return i18n_field


def get_translated_models(app_name):
    """
    Return models having a i18n = TranslationField() for given app_name.
    """
    app = apps.get_app_config(app_name)
    for model in app.get_models():
        i18n_field = get_i18n_field(model)
        if i18n_field is not None:
            yield model


def get_i18n_field_param(Model, i18n_field, param_name):
    """
    Return i18n_param from Model.i18n_field_params dict if exists
    or get param from i18n_field
    """
    if hasattr(Model, "i18n_field_params") and param_name in Model.i18n_field_params:
        return Model.i18n_field_params[param_name]
    return getattr(i18n_field, param_name)


def translate_model(Model):
    i18n_field = get_i18n_field(Model)

    if i18n_field is None:
        return

    if not i18n_field.virtual_fields:
        # This mode is required for the migration process:
        # It needs to have a stage where we do have the TranslationField,
        # but not the virtual fields (which would collide with the
        # django-modeltranslation `<field>_<lang>`-fields), to be able to
        # copy the values from the `<field>_<lang>`-fields into `i18n.<field>_<lang>`.
        return

    validate(Model)

    add_manager(Model)
    fields_to_translate = get_i18n_field_param(Model, i18n_field, "fields")
    required_languages = get_i18n_field_param(Model, i18n_field, "required_languages")
    add_virtual_fields(Model, fields_to_translate, required_languages)
    patch_constructor(Model)

    translate_meta_ordering(Model)


def check_languages(languages, model):
    for lang in languages:
        if lang not in get_available_languages():
            raise ImproperlyConfigured(
                'Language "{}" is in required_languages on Model "{}" but '
                "not in settings.MODELTRANS_AVAILABLE_LANGUAGES.".format(lang, model.__name__)
            )


def validate(Model):
    """
    Perform validation of the arguments to TranslationField
    """
    i18n_field = Model._meta.get_field("i18n")
    for field in i18n_field.fields:
        try:
            Model._meta.get_field(field)
        except FieldDoesNotExist:
            raise ImproperlyConfigured(
                'Argument "fields" to TranslationField contains an item "{}", '
                "which is not a field (missing a comma?).".format(field)
            )

    if i18n_field.fallback_language_field:
        field = get_model_field(Model, i18n_field.fallback_language_field)
        if field is None:
            raise ImproperlyConfigured(
                'Argument "fallback_language_field" to TranslationField is "{}", '
                "which is not an existing field.".format(i18n_field.fallback_language_field)
            )

    if i18n_field.required_languages:
        required_languages = i18n_field.required_languages
        allowed_types = (tuple, list, set)
        if isinstance(required_languages, allowed_types):
            check_languages(required_languages, Model)
        elif isinstance(required_languages, dict):
            for fieldname, languages in required_languages.items():
                if fieldname not in i18n_field.fields:
                    raise ImproperlyConfigured(
                        'Fieldname "{}" in required_languages which is not '
                        'defined as translatable in "{}.i18n".'.format(field, Model.__name__)
                    )
                if not isinstance(languages, allowed_types):
                    raise ImproperlyConfigured(
                        'required_languages["{}"] must be a tuple, list or set'.format(fieldname)
                    )
                check_languages(languages, Model)
        else:
            raise ImproperlyConfigured('"required_languages" must be a tuple, list, set or dict')


def raise_if_field_exists(Model, field_name):
    if not hasattr(Model, field_name):
        return

    try:
        Model._meta.get_field(field_name)
    except FieldDoesNotExist:
        return

    raise ImproperlyConfigured(
        'Error adding translation field. Model "{}" already contains '
        'a field named "{}".'.format(Model._meta.object_name, field_name)
    )


def add_virtual_fields(Model, fields, required_languages):
    """
    Adds newly created translation fields to the given translation options.
    """
    # Add virtual/proxy fields to assign values to and get values from.
    for field_name in fields:
        # if required_languages is dict, it contains an iterable of required languages
        # for each field.
        if isinstance(required_languages, dict):
            field_required_languages = required_languages.get(field_name, ())
        else:
            field_required_languages = required_languages

        original_field = Model._meta.get_field(field_name)

        # first, add a `<original_field_name>_i18n` virtual field to get the currently
        # active translation for a field
        field = translated_field_factory(
            original_field=original_field, blank=True, null=True, editable=False  # disable in admin
        )

        raise_if_field_exists(Model, field.get_field_name())
        field.contribute_to_class(Model, field.get_field_name())

        # add a virtual field pointing to the original field with name
        # <original_field_name>_<LANGUAGE_CODE>
        field = translated_field_factory(
            original_field=original_field,
            language=get_default_language(),
            blank=True,
            null=True,
            editable=False,
        )
        raise_if_field_exists(Model, field.get_field_name())
        field.contribute_to_class(Model, field.get_field_name())

        # now, for each language, add a virtual field to get the tranlation for
        # that specific langauge
        # <original_field_name>_<language>
        for language in get_available_languages(include_default=False):
            blank_allowed = language not in field_required_languages
            field = translated_field_factory(
                original_field=original_field,
                language=language,
                blank=blank_allowed,
                null=blank_allowed,
            )
            raise_if_field_exists(Model, field.get_field_name())
            field.contribute_to_class(Model, field.get_field_name())


def has_custom_queryset(manager):
    """
    Check whether manager (or its parents) has declared some custom get_queryset method.
    """
    return getattr(manager, "get_queryset", None) != getattr(Manager, "get_queryset", None)


def add_manager(model):
    """
    Monkey patches the original model to use MultilingualManager instead of
    default managers (not only ``objects``, but also every manager defined and inherited).

    Custom managers are merged with MultilingualManager.
    """
    if model._meta.abstract:
        return

    def patch_manager_class(manager):
        if isinstance(manager, MultilingualManager):
            return
        if manager.__class__ is Manager:
            manager.__class__ = MultilingualManager
        else:

            class NewMultilingualManager(MultilingualManager, manager.__class__):
                use_for_related_fields = getattr(
                    manager.__class__, "use_for_related_fields", not has_custom_queryset(manager)
                )
                _old_module = manager.__module__
                _old_class = manager.__class__.__name__

                def deconstruct(self):
                    return (
                        False,  # as_manager
                        "%s.%s" % (self._old_module, self._old_class),  # manager_class
                        None,  # qs_class
                        self._constructor_args[0],  # args
                        self._constructor_args[1],  # kwargs
                    )

            manager.__class__ = NewMultilingualManager

    # Make all managers local for this model to fix patching parent model managers
    model._meta.local_managers = model._meta.managers

    managers = model._meta.local_managers
    for current_manager in managers:
        prev_class = current_manager.__class__
        patch_manager_class(current_manager)
        if model._default_manager.__class__ is prev_class:
            # Normally model._default_manager is a reference to one of model's managers
            # (and would be patched by the way).
            # However, in some rare situations (mostly proxy models)
            # model._default_manager is not the same instance as one of managers, but it
            # share the same class.
            model._default_manager.__class__ = current_manager.__class__
    patch_manager_class(model._base_manager)
    if hasattr(model._meta, "_expire_cache"):
        model._meta._expire_cache()


def patch_constructor(model):
    """
    Monkey patches the original model to rewrite fields names in __init__
    """
    old_init = model.__init__

    def patched_init(self, *args, **kwargs):
        old_init(self, *args, **transform_translatable_fields(self.__class__, kwargs))

    model.__init__ = patched_init


def translate_meta_ordering(Model):
    """
    If a model has ``Meta.ordering`` defined, we check if
    one of it's fields is a translated field. If that's the case,
    add the expression to get the value from the i18n-field.
    """
    if len(Model._meta.ordering) == 0:
        return
    queryset = Model.objects.get_queryset()

    Model._meta.ordering = queryset._rewrite_ordering(Model._meta.ordering)



from django.db import connection
from django.test import TestCase
from django.utils.translation import override

from migrate_test.app.models import Blog


class PostMigrateTest(TestCase):
    def test_verify_installed_apps(self):
        from django.conf import settings

        self.assertIn("modeltrans", settings.INSTALLED_APPS)
        self.assertNotIn("modeltranslation", settings.INSTALLED_APPS)

    def test_model_fields(self):
        falcon = Blog.objects.get(title="Falcon")
        self.assertEquals(falcon.i18n["title_nl"], "Valk")
        self.assertEquals(falcon.i18n["title_de"], "Falk")
        self.assertIn("body_nl", falcon.i18n)

        with override("nl"):
            self.assertEquals(falcon.title_i18n, "Valk")

        with override("de"):
            self.assertEquals(falcon.title_i18n, "Falk")

    def test_indexes_in_place(self):
        """
        Check if the i18n column has the gin index.
        """
        db_table = Blog._meta.db_table
        index_prefix = "{}_i18n".format(db_table)

        with connection.cursor() as cursor:
            cursor.execute(
                "SELECT indexname, indexdef FROM pg_indexes WHERE tablename = %s;", [db_table]
            )
            indexes = {name: definition for name, definition in cursor.fetchall()}

        for name, definition in indexes.items():
            if name.startswith(index_prefix):
                self.assertIn("_gin", name)
                self.assertIn("USING gin", definition)



from django.contrib.postgres.fields.jsonb import KeyTransform
from django.core.exceptions import FieldDoesNotExist
from django.db.models.constants import LOOKUP_SEP
from django.db.models.lookups import Transform
from django.utils.translation import get_language as _get_language

from .conf import get_available_languages, get_default_language


def get_language():
    """
    Return an active language code that is guaranteed to be in settings.LANGUAGES

    (Django does not seem to guarantee this for us.)
    """
    lang = _get_language()
    if lang in get_available_languages():
        return lang
    return get_default_language()


def split_translated_fieldname(field_name):
    _pos = field_name.rfind("_")
    return (field_name[0:_pos], field_name[_pos + 1 :])


def build_localized_fieldname(field_name, lang):
    if lang == "id":
        # The 2-letter Indonesian language code is problematic with the
        # current naming scheme as Django foreign keys also add "id" suffix.
        lang = "ind"
    return "{}_{}".format(field_name, lang.replace("-", "_"))


def get_model_field(model, path):
    """
    Return the django model field for model in context, following relations.
    """

    if not hasattr(model, "_meta"):
        raise ValueError("Argument 'Model' must be a django.db.models.Model instance.")

    field = None
    for bit in path.split(LOOKUP_SEP):
        try:
            field = model._meta.get_field(bit)
        except FieldDoesNotExist:
            return None

        if hasattr(field, "remote_field"):
            rel = getattr(field, "remote_field", None)
            model = getattr(rel, "model", model)

    return field


def get_instance_field_value(instance, path):
    """
    Return the value of a field or related field for a model instance.
    """
    value = instance
    for bit in path.split(LOOKUP_SEP):
        try:
            value = getattr(value, bit)
        except AttributeError:
            return None

    return value


class FallbackTransform(Transform):
    """
    Custom version of KeyTextTransform to use a database field as part of the key.

    For example: with default_language="nl", calling
    `FallbackTransform("title_", F("fallback_language"), "i18n")` becomes in SQL"
        `"i18n"->>('title_' || "fallback_language")`
    """

    def __init__(self, field_prefix, language_expression, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.field_prefix = field_prefix
        self.language_expression = language_expression

    def preprocess_lhs(self, compiler, connection, lhs_only=False):
        if not lhs_only:
            key_transforms = [self.field_prefix]
        previous = self.lhs
        while isinstance(previous, KeyTransform):
            if not lhs_only:
                key_transforms.insert(0, previous.key_name)
            previous = previous.lhs
        lhs, params = compiler.compile(previous)
        return (lhs, params, key_transforms) if not lhs_only else (lhs, params)

    def as_postgresql(self, compiler, connection):
        lhs, params, key_transforms = self.preprocess_lhs(compiler, connection)
        params.extend([self.field_prefix])

        rhs = self.language_expression.resolve_expression(compiler.query)
        rhs_sql, rhs_params = compiler.compile(rhs)
        params.extend(rhs_params)

        return "(%s ->> (%%s || %s ))" % (lhs, rhs_sql), (params)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-16 15:21
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("app", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="blog", name="body_de", field=models.TextField(blank=True, null=True)
        ),
        migrations.AddField(
            model_name="blog", name="body_en", field=models.TextField(blank=True, null=True)
        ),
        migrations.AddField(
            model_name="blog", name="body_fr", field=models.TextField(blank=True, null=True)
        ),
        migrations.AddField(
            model_name="blog", name="body_nl", field=models.TextField(blank=True, null=True)
        ),
        migrations.AddField(
            model_name="blog", name="title_de", field=models.CharField(max_length=255, null=True)
        ),
        migrations.AddField(
            model_name="blog", name="title_en", field=models.CharField(max_length=255, null=True)
        ),
        migrations.AddField(
            model_name="blog", name="title_fr", field=models.CharField(max_length=255, null=True)
        ),
        migrations.AddField(
            model_name="blog", name="title_nl", field=models.CharField(max_length=255, null=True)
        ),
        migrations.AddField(
            model_name="category", name="name_de", field=models.CharField(max_length=255, null=True)
        ),
        migrations.AddField(
            model_name="category", name="name_en", field=models.CharField(max_length=255, null=True)
        ),
        migrations.AddField(
            model_name="category", name="name_fr", field=models.CharField(max_length=255, null=True)
        ),
        migrations.AddField(
            model_name="category", name="name_nl", field=models.CharField(max_length=255, null=True)
        ),
    ]



from django.core.exceptions import FieldDoesNotExist
from django.db.models import Count, Func, Manager, Q, QuerySet
from django.db.models.constants import LOOKUP_SEP
from django.db.models.expressions import CombinedExpression, F, OrderBy
from django.db.models.functions import Cast

from .conf import get_default_language
from .fields import TranslatedVirtualField


def transform_translatable_fields(model, fields):
    """
    Transform the kwargs for a <Model>.objects.create() or <Model>()
    to allow passing translated field names.

    Arguments:
        fields (dict): kwargs to a model __init__ or Model.objects.create() method
            for which the field names need to be translated to values in the i18n field
    """
    # If the current model does have the TranslationField, we must not apply
    # any transformation for it will result in a:
    # TypeError: 'i18n' is an invalid keyword argument for this function
    if not hasattr(model, "i18n"):
        return fields

    ret = {"i18n": fields.pop("i18n", None) or {}}

    # keep track of translated fields, and do not return an `i18n` key if no
    # translated fields are found.
    has_translated_fields = len(ret["i18n"].items()) > 0

    for field_name, value in fields.items():
        try:
            field = model._meta.get_field(field_name)
        except FieldDoesNotExist:
            ret[field_name] = value
            continue

        if isinstance(field, TranslatedVirtualField):
            has_translated_fields = True
            if field.get_language() == get_default_language():
                if field.original_name in fields:
                    raise ValueError(
                        'Attempted override of "{}" with "{}". '
                        "Only one of the two is allowed.".format(field.original_name, field_name)
                    )
                ret[field.original_name] = value
            else:
                ret["i18n"][field.name] = value
        else:
            ret[field_name] = value

    if not has_translated_fields:
        return fields

    return ret


class MultilingualQuerySet(QuerySet):
    """
    Extends ``~django.db.models.query.QuerySet`` and makes the translated versions of fields
    accessible through the normal QuerySet methods, analogous to the virtual fields added
    to a translated model:

     - `<field>` allow getting/setting the default language
     - ``<field>_<lang>`` (for example, `<field>_de`) allows getting/setting a specific language.
       Note that if `LANGUAGE_CODE == "en"`, `<field>_en` is mapped to `<field>`.
     - `<field>_i18n` follows the currently active translation in Django, and falls back to the default language.

    When adding the `modeltrans.fields.TranslationField` to a model, MultilingualManager is automatically
    mixed in to the manager class of that model.
    """

    def _add_i18n_annotation(
        self, virtual_field=None, fallback=True, bare_lookup=None, annotation_name=None
    ):
        """
        Private method to add an annotation to the query to extract the translated
        version of a field from the jsonb field to allow filtering and ordering.

        Arguments:
            virtual_field (TranslatedVirtualField): the virtual field to create an annotation for.
            fallback (bool): If `True`, `COALESCE` will be used to get the value of the original
                field if the requested translation is not in the `i18n` dict.
            bare_lookup:
            annotation_name (str): name of the annotation, if None the default
                `<original_field>_<lang>_annotation` will be used.

        Returns:
            the name of the annotation created.
        """
        expression = virtual_field.as_expression(fallback=fallback, bare_lookup=bare_lookup)

        if isinstance(expression, F):
            return expression.name

        if annotation_name is None:
            annotation_name = "{}_annotation".format(virtual_field.name)

        self.query.add_annotation(expression, annotation_name)
        return annotation_name

    def _get_field(self, lookup):
        """
        Return the Django model field for a lookup plus the remainder of the lookup,
        which should be the lookup type.
        """
        model = self.model
        lookup_type = None

        # pk is not an actual field, but an alias for the implicit id field.
        if lookup == "pk":
            key = None
            for field in model._meta.get_fields():
                if getattr(field, "primary_key", False):
                    key = field
            return key, None

        field = None
        bits = lookup.split(LOOKUP_SEP)

        for i, bit in enumerate(bits):
            try:
                field = model._meta.get_field(bit)
            except FieldDoesNotExist:
                lookup_type = LOOKUP_SEP.join(bits[i:])
                break

            if hasattr(field, "remote_field"):
                rel = getattr(field, "remote_field", None)
                model = getattr(rel, "model", model)

        return field, lookup_type

    def _rewrite_filter_clause(self, lookup, value):
        """
        Rewrite a filter clause passed to filter()/exclude()/etc.

        for example:

        for title_nl__like="va"
        _rewrite_filter_clause("title_nl__like", "va") should be called.
        """
        value = self._rewrite_expression(value)
        field, lookup_type = self._get_field(lookup)

        if not isinstance(field, TranslatedVirtualField):
            return lookup, value

        if lookup_type is not None:
            bare_lookup = lookup[0 : -(len(LOOKUP_SEP + lookup_type))]
        else:
            bare_lookup = lookup

        filter_field_name = self._add_i18n_annotation(
            virtual_field=field, bare_lookup=bare_lookup, fallback=field.language is None
        )

        # re-add lookup type
        if lookup_type is not None:
            filter_field_name += LOOKUP_SEP + lookup_type

        return filter_field_name, value

    def _rewrite_expression(self, expr):
        """
        Rewrite expressions.

        https://docs.djangoproject.com/en/stable/ref/models/expressions/

        This current way of doing this is bound to lag behind any new things implemented in Django.
        It would be really nice to have a better/more generic way of doing this.
        """
        if isinstance(expr, F):
            field, _ = self._get_field(expr.name)
            if not isinstance(field, TranslatedVirtualField):
                return expr

            return field.as_expression(fallback=field.language is None, bare_lookup=expr.name)
        elif isinstance(expr, CombinedExpression):
            expr.lhs = self._rewrite_expression(expr.lhs)
            expr.rhs = self._rewrite_expression(expr.rhs)
        elif isinstance(expr, Count):
            expr.source_expressions[0] = self._rewrite_expression(expr.source_expressions[0])
        elif isinstance(expr, Func):
            expr.source_expressions = list(
                [self._rewrite_expression(e) for e in expr.source_expressions]
            )
        elif isinstance(expr, OrderBy):
            expr.expression = self._rewrite_expression(expr.expression)
        return expr

    def _rewrite_Q(self, q):
        if isinstance(q, Q):
            return Q._new_instance(
                list(self._rewrite_Q(child) for child in q.children),
                connector=q.connector,
                negated=q.negated,
            )
        if isinstance(q, (list, tuple)):
            return self._rewrite_filter_clause(*q)

    def _rewrite_ordering(self, field_names):
        new_field_names = []

        for field_name in field_names:
            if not isinstance(field_name, str):
                new_field_names.append(self._rewrite_expression(field_name))
                continue

            # remove descending prefix, not relevant for the annotation
            sort_order = ""
            if field_name[0] == "-":
                field_name = field_name[1:]
                sort_order = "-"

            if field_name == "pk":
                new_field_names.append(sort_order + "pk")
                continue

            field, lookup_type = self._get_field(field_name)
            if field is None or not isinstance(field, TranslatedVirtualField):
                # if the field is just a normal field or not a field
                # no rewriting needed
                new_field_names.append(sort_order + field_name)
                continue

            assert lookup_type is None, "{} is not a valid order_by lookup".format(field_name)

            sort_field = field.as_expression(bare_lookup=field_name)
            if sort_order == "-":
                sort_field = sort_field.desc()

            new_field_names.append(sort_field)

        return new_field_names

    def annotate(self, *args, **kwargs):
        """
        Patch annotate to allow the use of translated field names in annotations.

        https://docs.djangoproject.com/en/stable/ref/models/querysets/#annotate
        """
        args = [self._rewrite_expression(a) for a in args]
        kwargs = {alias: self._rewrite_expression(expr) for alias, expr in kwargs.items()}

        return super().annotate(*args, **kwargs)

    def create(self, **kwargs):
        """
        Patch the create method to allow adding the value for a translated field
        using `Model.objects.create(..., title_nl="...")`.

        https://docs.djangoproject.com/en/stable/ref/models/querysets/#create
        """
        return super().create(**transform_translatable_fields(self.model, kwargs))

    def order_by(self, *field_names):
        """
        Annotate translated fields before sorting.

        Examples:
         - sort on `-title_nl` will add an annotation for `title_nl`
         - sort on `title_i18n` will add an annotation for the current language

        The field names pointing to translated fields in the `field_names`
        argument will be replaced by their annotated versions.

        https://docs.djangoproject.com/en/1.11/ref/models/querysets/#order_by
        """

        new_field_names = self._rewrite_ordering(field_names)

        return super().order_by(*new_field_names)

    def _filter_or_exclude(self, negate, *args, **kwargs):
        """
        Annotate lookups for `filter()` and `exclude()`.

        Examples:
            - `title_nl__contains="foo"` will add an annotation for `title_nl`
            - `title_nl="bar"` will add an annotation for `title_nl`
            - `title_i18n="foo"` will add an annotation for a coalesce of the
               current active language, and all items of the fallback chain.
            - `Q(title_nl__contains="foo") will add an annotation for `title_nl`

        In all cases, the field part of the field lookup will be changed to use
        the annotated verion.
        """
        # handle Q expressions / args
        new_args = []
        for arg in args:
            new_args.append(Q(self._rewrite_Q(arg)))

        # handle the kwargs
        new_kwargs = {}
        for field, value in kwargs.items():
            new_kwargs.update(dict((self._rewrite_filter_clause(field, value),)))

        return super()._filter_or_exclude(negate, *new_args, **new_kwargs)

    def _values(self, *fields, **expressions):
        """
        Annotate lookups for `values()` and `values_list()`

        It must be possible to use:
        `Blogs.objects.all().values_list("title_i18n", "title_nl", "title_en")`

        But also spanning relations:
        `Blogs.objects.all().values_list("title_i18n", "category__name__i18n")`
        """
        _fields = fields + tuple(expressions)

        for field_name in _fields:
            field, lookup_type = self._get_field(field_name)
            if not isinstance(field, TranslatedVirtualField):
                continue

            fallback = field.language is None

            if field.get_language() == get_default_language():
                original_field = field_name.replace(field.name, field.original_field.name)
                self.query.add_annotation(Cast(original_field, field.output_field()), field_name)
            else:
                self._add_i18n_annotation(
                    virtual_field=field,
                    fallback=fallback,
                    bare_lookup=field_name,
                    annotation_name=field_name,
                )

        return super()._values(*fields, **expressions)

    def __reduce__(self):
        """
        Make sure a dynamic version of this class can be pickled
        """
        return multilingual_queryset_factory, (self.__class__.__bases__[0],), self.__getstate__()


def multilingual_queryset_factory(old_cls, instantiate=True):
    """Return a MultilingualQuerySet, or mix MultilingualQuerySet in custom QuerySets."""
    if old_cls == QuerySet:
        NewClass = MultilingualQuerySet
    else:

        class NewClass(old_cls, MultilingualQuerySet):
            pass

        NewClass.__name__ = "Multilingual%s" % old_cls.__name__
    return NewClass() if instantiate else NewClass


class MultilingualManager(Manager):
    """
    When adding the `modeltrans.fields.TranslationField` to a model, MultilingualManager is automatically
    mixed in to the manager class of that model.

    If you want to use translated fields when building the query from a related model, you need to add
    ``objects = MultilingualManager()`` to the model you want to build the query from.

    For example, ``Category`` needs ``objects = MultilingualManager()`` in order to allow
    ``Category.objects.filter(blog__title_i18n__icontains="django")``::

        class Category(models.Model):
            title = models.CharField(max_length=255)

            objects = MultilingualManager()  # required to use translated fields of Blog.

        class Blog(models.Model):
            title = models.CharField(max_length=255)
            body = models.TextField(null=True)
            category = models.ForeignKey(Category, null=True, blank=True, on_delete=models.CASCADE)

            i18n = TranslationField(fields=("title", "body"))
    """

    use_for_related_fields = True

    def _patch_queryset(self, qs):
        qs.__class__ = multilingual_queryset_factory(qs.__class__, instantiate=False)
        return qs

    def get_queryset(self):
        """
        This method is repeated because some managers that don't use super() or alter the QuerySet class
        may return QuerySet that is not subclass of MultilingualQuerySet.
        """
        qs = super().get_queryset()
        if isinstance(qs, MultilingualQuerySet):
            # Is already patched
            return qs
        return self._patch_queryset(qs)



from collections import defaultdict

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Creates the i18n datamigration (django-modeltranslation -> django-modeltrans) for the specified app"

    def add_arguments(self, parser):
        parser.add_argument("apps", nargs="+", type=str)

    def handle(self, *args, **options):
        from modeltrans.migration import (
            I18nDataMigration,
            get_translatable_models,
            get_translated_fields,
        )

        models = get_translatable_models()

        apps = defaultdict(list)
        for model in models:
            apps[model._meta.app_label].append(model)

        for app in options["apps"]:
            print("Create data migration for app:", app)

            if len(apps[app]) == 0:
                print("No models registered for translation with django-modeltranslation")
                break
            migration = I18nDataMigration(app)

            for Model in apps[app]:
                translatable_fields = tuple(get_translated_fields(Model))

                print(
                    'added model "{}" with fields [{}]'.format(
                        Model.__name__, str(translatable_fields)
                    )
                )
                migration.add_model(Model, translatable_fields)

            filename = migration.write_migration_file()
            print("Wrote migration for {} to {}".format(app, filename))



import itertools

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


def get_modeltrans_setting(key):
    modeltrans_settings = {
        "MODELTRANS_AVAILABLE_LANGUAGES": get_available_languages(),
        "MODELTRANS_FALLBACK": getattr(
            settings, "MODELTRANS_FALLBACK", {"default": (get_default_language(),)}
        ),
        "MODELTRANS_ADD_FIELD_HELP_TEXT": getattr(settings, "MODELTRANS_ADD_FIELD_HELP_TEXT", True),
    }
    return modeltrans_settings.get(key)


def get_default_language():
    return settings.LANGUAGE_CODE


def get_available_languages_setting():
    """
    list of available languages for modeltrans translations.
    defaults to the list of language codes extracted from django setting LANGUAGES
    """
    languages = tuple(
        set(
            getattr(
                settings,
                "MODELTRANS_AVAILABLE_LANGUAGES",
                (code for code, _ in getattr(settings, "LANGUAGES")),
            )
        )
    )

    if not all(isinstance(x, str) for x in languages):
        raise ImproperlyConfigured(
            "MODELTRANS_AVAILABLE_LANGUAGES should be an iterable of strings"
        )

    # make sure LANGUAGE_CODE is not in available languages
    return (lang for lang in languages if lang != get_default_language())


def get_available_languages(include_default=True):
    """
    Returns a tuple of available languages for django-modeltrans.
    """
    MODELTRANS_AVAILABLE_LANGUAGES = get_available_languages_setting()

    if include_default:
        return tuple(
            set(itertools.chain(MODELTRANS_AVAILABLE_LANGUAGES, (get_default_language(),)))
        )
    else:
        return MODELTRANS_AVAILABLE_LANGUAGES


def check_fallback_chain():
    MODELTRANS_FALLBACK = get_modeltrans_setting("MODELTRANS_FALLBACK")
    MODELTRANS_AVAILABLE_LANGUAGES = get_available_languages(include_default=True)

    if "default" not in MODELTRANS_FALLBACK:
        raise ImproperlyConfigured("MODELTRANS_FALLBACK setting must have a `default` key.")

    message_fmt = (
        "MODELTRANS_FALLBACK contains language `{}` "
        "which is not in MODELTRANS_AVAILABLE_LANGUAGES"
    )
    for lang, chain in MODELTRANS_FALLBACK.items():
        if lang != "default" and lang not in MODELTRANS_AVAILABLE_LANGUAGES:
            raise ImproperlyConfigured(message_fmt.format(lang))
        for lang in chain:
            if lang not in MODELTRANS_AVAILABLE_LANGUAGES:
                raise ImproperlyConfigured(message_fmt.format(lang))


def get_fallback_chain(lang):
    """
    Returns the list of fallback languages for language `lang`.

    For example, this function will return `('nl', 'en')` when called
    with `lang='fy'` and configured like this::

        LANGUAGE_CODE = 'en'

        MODELTRANS_FALLBACK = {
           'default': (LANGUAGE_CODE, ),
           'fy': ('nl', 'en')
        }
    """
    MODELTRANS_FALLBACK = get_modeltrans_setting("MODELTRANS_FALLBACK")

    if lang not in MODELTRANS_FALLBACK.keys():
        lang = "default"

    return MODELTRANS_FALLBACK[lang]



#!/usr/bin/env python
import os
import re
import sys

from setuptools import find_packages, setup

# get version without importing
with open("modeltrans/__init__.py", "rb") as f:
    VERSION = str(re.search('__version__ = "(.+?)"', f.read().decode("utf-8")).group(1))

if sys.argv[-1] == "publish":
    os.system("python setup.py sdist")
    os.system("twine upload dist/django-modeltrans-{}.tar.gz".format(VERSION))
    message = "\nreleased [{version}](https://pypi.python.org/pypi/django-modeltrans/{version})"
    print(message.format(version=VERSION))
    sys.exit()

if sys.argv[-1] == "tag":
    os.system("git tag -a v{} -m 'tagging v{}'".format(VERSION, VERSION))
    os.system("git push --tags && git push origin master")
    sys.exit()


setup(
    name="django-modeltrans",
    version=VERSION,
    description="Model translations in a jsonb field",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Jan Pieter Waagmeester",
    author_email="jieter@zostera.nl",
    license="Simplified BSD",
    url="https://github.com/zostera/django-modeltrans/",
    packages=find_packages(exclude=["tests.*", "tests", "example.*", "example"]),
    include_package_data=True,  # declarations in MANIFEST.in
    install_requires=["Django>=1.11.15"],
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries",
    ],
)



import inspect
import os
import sys

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.migrations.loader import MigrationLoader
from django.utils.timezone import now

from . import __version__ as VERSION
from .conf import get_default_language
from .utils import split_translated_fieldname

try:
    from modeltranslation.translator import translator

    DJANGO_MODELTRANSLATION_AVAILABLE = True
except ImportError:
    DJANGO_MODELTRANSLATION_AVAILABLE = False

DEFAULT_LANGUAGE = get_default_language()


def _raise_if_not_django_modeltranslation():
    """Raise if we cannot import django-modeltranslation during migration"""
    if not DJANGO_MODELTRANSLATION_AVAILABLE:
        raise ImproperlyConfigured(
            "django-modeltranslation must be still installed when creating"
            "the modeltranslation -> modeltrans migrations."
        )


def get_translatable_models():
    """
    Get the translatable models according to django-modeltranslation

    !! only use to migrate from django-modeltranslation !!
    """
    _raise_if_not_django_modeltranslation()
    return translator.get_registered_models()


def get_translated_fields(Model):
    """
    Enumerates the translated fields for a model according to django-modeltranslation.
    For example: title_nl, title_en, title_fr, body_nl, body_en, body_fr

    !! only use to migrate from django-modeltranslation !!
    """
    _raise_if_not_django_modeltranslation()

    options = translator.get_options_for_model(Model)
    for original_field, fields in options.fields.items():
        for translated in fields:
            yield translated.name


def copy_translations(Model, fields):
    """
    Copy translations for all items in the database for a Model with
    translations managed by django-modeltranslation into a json field `i18n`
    managed by django-modeltrans.
    Values for the default language will be copied to the original field.

    Arguments:
        Model: A (historical) Model from the migraton's app registry
        fields(iterable): list of fields to copy into their new places.
    """
    for m in Model.objects.all():
        m.i18n = {}
        for field in fields:
            value = getattr(m, field)
            if value in (None, ""):
                continue

            original_field, lang = split_translated_fieldname(field)

            if lang == DEFAULT_LANGUAGE:
                setattr(m, original_field, value)
            else:
                m.i18n[field] = value

        m.save()


def get_latest_migration(app_name, connection=None):
    """
    Get the name of the latest applied migration and raises if unapplied
    migrations exist for the app.

    Arguments:
        app_name(str): Name of the app.
        connection: database connection to get the latest migration for.
    Simplified version of
    https://github.com/django/django/blob/1.9.2/django/core/management/commands/showmigrations.py#L38-L77
    """
    if connection is None:
        connection = connections[DEFAULT_DB_ALIAS]

    loader = MigrationLoader(connection, ignore_no_migrations=True)
    graph = loader.graph
    last = None
    shown = set()
    for node in graph.leaf_nodes(app_name):
        for plan_node in graph.forwards_plan(node):
            if plan_node not in shown and plan_node[0] == app_name:
                if plan_node in loader.applied_migrations:
                    last = plan_node[1]
                else:
                    raise Exception("You have unapplied migration(s) for app {}".format(app_name))
                shown.add(plan_node)

    return last


def get_next_migration_filename(app_name, connection=None, migration_type="data"):
    """
    Return name (including the absolute path) of the next migration to insert for this app
    """
    latest_migration_name = get_latest_migration(app_name)
    next_migration_name = "{0:04d}_i18n_{1}_migration.py".format(
        int(latest_migration_name[0:4]) + 1, migration_type
    )
    app_base_path = os.path.dirname(apps.get_app_config(app_name).module.__file__)

    return os.path.join(app_base_path, "migrations", next_migration_name)


class I18nMigration(object):
    helper_functions = ()
    template = """
# -*- coding: utf-8 -*-
# Generated by django-modeltrans {version} on {timestamp}

from __future__ import print_function, unicode_literals

from django.db import migrations

DEFAULT_LANGUAGE = "{DEFAULT_LANGUAGE}"

{helpers}


class Migration(migrations.Migration):

    dependencies = [
        ("{app}", "{last_migration}"),
    ]

    operations = [
        {operations}
    ]
"""

    def __init__(self, app):
        self.models = []
        self.app = app

        self.migration_filename = (
            get_latest_migration(self.app) or "# TODO: manually insert latest migration here"
        )

    def get_helper_functions(self):
        def to_str(fn):
            return inspect.getsource(fn) if callable(fn) else fn

        for fn in self.helper_functions:
            yield to_str(fn)

        for fn in self.get_extra_helper_functions():
            yield to_str(fn)

    def get_extra_helper_functions(self):
        return []

    def add_model(self, Model, fields):
        self.models.append((Model, fields))

    def get_helper_src(self):
        return "\n\n".join(self.get_helper_functions())

    def write(self, out=None):
        if out is None:
            out = sys.stdout

        out.write(
            self.template.format(
                version=VERSION,
                DEFAULT_LANGUAGE=getattr(
                    settings, "MODELTRANSLATION_DEFAULT_LANGUAGE", get_default_language()
                ),
                timestamp=now().strftime("%Y-%m-%d %H:%M"),
                helpers=self.get_helper_src(),
                app=self.app,
                last_migration=self.migration_filename,
                operations=self.get_operations(),
            )
        )

    def write_migration_file(self):
        """
        Write the migration to file.
        """
        filename = get_next_migration_filename(self.app, migration_type=self.migration_type)
        with open(filename, "w") as f:
            self.write(f)

        return filename


class I18nDataMigration(I18nMigration):
    migration_type = "data"
    helper_functions = (split_translated_fieldname, copy_translations)

    forwards_template = """
def forwards(apps, schema_editor):
    app = '{app}'
    todo = (
        {todo},
    )

    for model, fields in todo:
        Model = apps.get_model(app, model)

        copy_translations(Model, fields)
"""

    def get_extra_helper_functions(self):
        yield self.forwards_template.format(
            todo=",\n        ".join(
                [str((Model.__name__, fields)) for Model, fields in self.models]
            ),
            app=self.app,
        )

    def get_operations(self):
        return """
        # The copying of values is (sort of) reversable by a no-op:
        #  - values are copied into i18n (which is not used by anything but django-modeltrans)
        #  - the default language is copied to the orignal field, which was not used
        #    with django-modeltrans.
        migrations.RunPython(forwards, migrations.RunPython.noop),
"""



from django.core.exceptions import ImproperlyConfigured
from django.db.models import F, fields
from django.db.models.functions import Cast, Coalesce
from django.utils.translation import gettext_lazy as _

from .conf import get_default_language, get_fallback_chain, get_modeltrans_setting
from .utils import (
    FallbackTransform,
    build_localized_fieldname,
    get_instance_field_value,
    get_language,
)

try:
    from django.db.models import JSONField, KeyTextTransform  # django==3.1 moved json field
except ImportError:
    from django.contrib.postgres.fields import JSONField
    from django.contrib.postgres.fields.jsonb import KeyTextTransform


SUPPORTED_FIELDS = (fields.CharField, fields.TextField)

DEFAULT_LANGUAGE = get_default_language()


def translated_field_factory(original_field, language=None, *args, **kwargs):
    if not isinstance(original_field, SUPPORTED_FIELDS):
        raise ImproperlyConfigured(
            "{} is not supported by django-modeltrans.".format(original_field.__class__.__name__)
        )

    class Specific(TranslatedVirtualField, original_field.__class__):
        pass

    Specific.__name__ = "Translated{}".format(original_field.__class__.__name__)

    return Specific(original_field, language, *args, **kwargs)


class TranslatedVirtualField:
    """
    A field representing a single field translated to a specific language.

    Arguments:
        original_field: The original field to be translated
        language: The language to translate to, or `None` to track the current active Django language.
    """

    # Implementation inspired by HStoreVirtualMixin from:
    # https://github.com/djangonauts/django-hstore/blob/master/django_hstore/virtual.py

    def __init__(self, original_field, language=None, *args, **kwargs):
        # TODO: this feels like a big hack.
        self.__dict__.update(original_field.__dict__)

        self.original_field = original_field
        self.language = language

        self.blank = kwargs["blank"]
        self.null = kwargs["null"]

        self.concrete = False
        self._help_text = kwargs.pop("help_text", None)

    @property
    def original_name(self):
        return self.original_field.name

    @property
    def help_text(self):
        if self._help_text is not None:
            return self._help_text

        if get_modeltrans_setting("MODELTRANS_ADD_FIELD_HELP_TEXT") and self.language is None:
            return _("current language: {}").format(get_language())

    def contribute_to_class(self, cls, name):
        self.model = cls

        self.attname = name
        self.name = name
        self.column = None

        # Use a translated verbose name:
        translated_field_name = _(self.original_field.verbose_name)
        if self.language is not None:
            translated_field_name += " ({})".format(self.language.upper())
        self.verbose_name = translated_field_name

        setattr(cls, name, self)
        cls._meta.add_field(self, private=True)

    def db_type(self, connection):
        return None

    def get_instance_fallback_chain(self, instance, language):
        """
        Return the fallback chain for the instance.

        Most of the time, it is just the configured fallback chain, but if the per-record-fallback feature
        is used, the value of the field is added (if not None).
        """
        default = get_fallback_chain(language)

        i18n_field = instance._meta.get_field("i18n")
        if i18n_field.fallback_language_field:
            record_fallback_language = get_instance_field_value(
                instance, i18n_field.fallback_language_field
            )

            if record_fallback_language:
                return (record_fallback_language, *default)

        return default

    def __get__(self, instance, instance_type=None):
        # This method is apparantly called with instance=None from django.
        # django-hstor raises AttributeError here, but that doesn't solve our problem.
        if instance is None:
            return

        if "i18n" in instance.get_deferred_fields():
            raise ValueError(
                "Getting translated values on a model fetched with defer('i18n') is not supported."
            )

        language = self.get_language()
        original_value = getattr(instance, self.original_name)
        if language == DEFAULT_LANGUAGE and original_value:
            return original_value

        # Make sure we test for containment in a dict, not in None
        if instance.i18n is None:
            instance.i18n = {}

        field_name = build_localized_fieldname(self.original_name, language)

        # Just return the value if this is an explicit field (<name>_<lang>)
        if self.language is not None:
            return instance.i18n.get(field_name)

        # This is the _i18n version of the field, and the current language is not available,
        # so we walk the fallback chain:
        for fallback_language in (language,) + self.get_instance_fallback_chain(instance, language):
            if fallback_language == DEFAULT_LANGUAGE:
                if original_value:
                    return original_value
                else:
                    continue

            field_name = build_localized_fieldname(self.original_name, fallback_language)
            if field_name in instance.i18n and instance.i18n[field_name]:
                return instance.i18n.get(field_name)

        # finally, return the original field if all else fails.
        return getattr(instance, self.original_name)

    def __set__(self, instance, value):
        if instance.i18n is None:
            instance.i18n = {}

        language = self.get_language()

        if language == DEFAULT_LANGUAGE:
            setattr(instance, self.original_name, value)
        else:
            field_name = build_localized_fieldname(self.original_name, language)

            # if value is None, remove field from `i18n`.
            if value is None:
                instance.i18n.pop(field_name, None)
            else:
                instance.i18n[field_name] = value

    def get_field_name(self):
        """
        Returns the field name for the current virtual field.

        The field name is ``<original_field_name>_<language>`` in case of a specific
        translation or ``<original_field_name>_i18n`` for the currently active language.
        """
        if self.language is None:
            lang = "i18n"
        else:
            lang = self.get_language()

        return build_localized_fieldname(self.original_name, lang)

    def get_language(self):
        """
        Returns the language for this field.

        In case of an explicit language (title_en), it returns "en", in case of
        `title_i18n`, it returns the currently active Django language.
        """
        return self.language if self.language is not None else get_language()

    def output_field(self):
        """
        The type of field used to Cast/Coalesce to.

        Mainly because a max_length argument is required for CharField
        until this PR is merged: https://github.com/django/django/pull/8758
        """
        Field = self.original_field.__class__
        if isinstance(self.original_field, fields.CharField):
            return Field(max_length=self.original_field.max_length)

        return Field()

    def _localized_lookup(self, language, bare_lookup):
        if language == DEFAULT_LANGUAGE:
            return bare_lookup.replace(self.name, self.original_name)

        # When accessing a table directly, the i18_lookup will be just "i18n", while following relations
        # they are in the lookup first.
        i18n_lookup = bare_lookup.replace(self.name, "i18n")

        # To support per-row fallback languages, an F-expression is passed as language parameter.
        if isinstance(language, F):
            # abuse build_localized_fieldname without language to get "<field>_"
            field_prefix = build_localized_fieldname(self.original_name, "")
            return FallbackTransform(field_prefix, language, i18n_lookup)
        else:
            return KeyTextTransform(
                build_localized_fieldname(self.original_name, language), i18n_lookup
            )

    def as_expression(self, bare_lookup, fallback=True):
        """
        Compose an expression to get the value for this virtual field in a query.
        """
        language = self.get_language()
        if language == DEFAULT_LANGUAGE:
            return F(self._localized_lookup(language, bare_lookup))

        if not fallback:
            i18n_lookup = self._localized_lookup(language, bare_lookup)
            return Cast(i18n_lookup, self.output_field())

        fallback_chain = get_fallback_chain(language)
        # First, add the current language to the list of lookups
        lookups = [self._localized_lookup(language, bare_lookup)]

        # Optionnally add the lookup for the per-row fallback language
        i18n_field = self.model._meta.get_field("i18n")
        if i18n_field.fallback_language_field:
            lookups.append(
                self._localized_lookup(F(i18n_field.fallback_language_field), bare_lookup)
            )

        # and now, add the list of fallback languages to the lookup list
        for fallback_language in fallback_chain:
            lookups.append(self._localized_lookup(fallback_language, bare_lookup))
        return Coalesce(*lookups, output_field=self.output_field())


class TranslationField(JSONField):
    """
    This model field is used to store the translations in the translated model.

    Arguments:
        fields (iterable): List of model field names to make translatable.
        required_languages (iterable or dict): List of languages required for the model.
            If a dict is supplied, the keys must be translated field names with the value
            containing a list of required languages for that specific field.
        virtual_fields (bool): If `False`, do not add virtual fields to access
            translated values with.
            Set to `True` during migration from django-modeltranslation to prevent
            collisions with it's database fields while having the `i18n` field available.
        fallback_language_field: If not None, this should be the name of the field containing a
            language code to use as the first language in any fallback chain.
            For example: if you have a model instance with 'nl' as language_code, and set
            fallback_language_field='language_code', 'nl' will always be tried after the current
            language before any other language.
    """

    description = "Translation storage for a model"

    def __init__(
        self,
        fields=None,
        required_languages=None,
        virtual_fields=True,
        fallback_language_field=None,
        *args,
        **kwargs,
    ):
        self.fields = fields or ()
        self.required_languages = required_languages or ()
        self.virtual_fields = virtual_fields
        self.fallback_language_field = fallback_language_field

        kwargs["editable"] = False
        kwargs["null"] = True
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        del kwargs["editable"]
        del kwargs["null"]
        kwargs["fields"] = self.fields
        kwargs["required_languages"] = self.required_languages
        kwargs["virtual_fields"] = self.virtual_fields

        return name, path, args, kwargs

    def get_translated_fields(self):
        """Return a generator for all translated fields."""
        for field in self.model._meta.get_fields():
            if isinstance(field, TranslatedVirtualField):
                yield field

    def contribute_to_class(self, cls, name):
        if name != "i18n":
            raise ImproperlyConfigured('{} must have name "i18n"'.format(self.__class__.__name__))

        super().contribute_to_class(cls, name)



#!/usr/bin/env python
from __future__ import print_function

import os
import sys
from subprocess import STDOUT, CalledProcessError, check_output


def cmd(c):
    print("\033[92m Running command: \033[0m", c)
    try:
        return check_output(c, shell=True, stderr=STDOUT)
    except CalledProcessError as e:
        print("\033[31m Process errored: \033[0m, code: {}".format(e.returncode))
        print(e.output)
        sys.exit(1)


os.chdir(os.path.dirname(os.path.abspath(__file__)))
cmd("git clean migrate_test/ -f")
cmd("git checkout -- migrate_test/")
